function rbss(n)
% rbss (Companion Code)
% by Christian Himpe, 2014 ( http://gramian.de )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%*

if(exist('emgr')~=2) disp('emgr framework is required. Download at http://gramian.de/emgr.m'); return; end

rand('seed',1009);
randn('seed',1009);

%%%%%%%% Setup %%%%%%%%

if(nargin<1), n = 4; end;

J = 1;
N = 100;
O = 1;
M = 100;

S = 0;
h = 0.01;
T = 1.0;

%[A,B,C] = ilp(J,N,O,1,1009); 

A = full(sprand(N,N,0.2));
%A = 0.5*(A+A');
B = rand(N,J);
C = rand(O,N);

X = zeros(N,1);
U = [ones(J,1)./h,zeros(J,(T/h)-1)];

P = -randn(N,1);
A(1:N+1:end) = -0.45*N*ones(N,1);

STABLE = all(eig(A+diag(P))<0)

Fl = @(x,u,p) (A+diag(p))*x + B*u;
Fa = @(x,u,p) (A+diag(p))'*x + C'*u;
Gl = @(x,u,p) C*x;
Ga = @(x,u,p) B'*x;
Gu = @(x,u,p) x;

l2norm = @(m) sqrt(h*sum(sum(m.*m,1)));

%tic;
Y = assess(Fl,Gl,X,U,P,S,h,T,speye(N),speye(N),speye(N),speye(N));
%toc;


%plot(h:h:T,Yl'); return;

L2 = zeros(9,N);
l2 = zeros(n-1,n-1,3);

%%%%%%%% Offline %%%%%%%%

tic;
WU = [assess(Fl,Gu,X,0.25*U,P,S,h,T,speye(N),speye(N),speye(N),speye(N)),...
      assess(Fl,Gu,X,0.50*U,P,S,h,T,speye(N),speye(N),speye(N),speye(N)),...
      assess(Fl,Gu,X,0.75*U,P,S,h,T,speye(N),speye(N),speye(N),speye(N)),...
      assess(Fl,Gu,X,1.00*U,P,S,h,T,speye(N),speye(N),speye(N),speye(N))];
off_wu = toc; fprintf('#');

tic;
WC = emgr(Fl,Gl,[J N O],[S h T],'c',P,0,U);
off_wc = toc; fprintf('#');

tic;
WO = emgr(Fl,Gl,[J N O],[S h T],'o',P,0,U);
off_wo = toc; fprintf('#');

tic;
WQ = emgr(Fa,Ga,[J N O],[S h T],'c',P,0,U);
off_wq = toc; fprintf('#');

tic;
WX = emgr(Fl,Gl,[J N O],[S h T],'x',P,0,U);
off_wx = toc; fprintf('#');

tic;
WY = emgr(Fl,Fa,[J N O],[S h T],'y',P,0,U);
off_wy = toc; fprintf('#');

tic;
WS = emgr(Fl,Gl,[J N O],[S h T],'s',P,[0,0,0,0,0,0,0,0,0,1,0,0],U,1);
off_ws = toc; fprintf('#');

tic;
WI = emgr(Fl,Gl,[J N O],[S h T],'i',P,0,U,0,1);
off_wi = toc; fprintf('#');

tic;
WJ = emgr(Fl,Gl,[J N O],[S h T],'j',P,0,U,0,1);
off_wj = toc; fprintf('#\n');


%%%%%%%% Offline II %%%%%%%%

tic;
[U1,D1,V1] = svd(WU);
off_pod = toc; fprintf('#');

tic;
[U2,D2,V2] = svd(WC);
off_wcc = toc; fprintf('#');

tic;
[U3,D3,V3] = bal(WC,WO);
off_bt1 = toc; fprintf('#');

tic;
[U4,D4,V4] = bal(WC,WQ);
off_bt2 = toc; fprintf('#');

tic;
[U5,D5,V5] = svd(WX);
off_wx1 = toc; fprintf('#');

tic;
[U6,D6,V6] = svd(WY);
off_wx2 = toc; fprintf('#');

tic;
[P1,E1,Q1] = svd(WS{2});
off_wsp = toc; fprintf('#');

tic;
[P2,E2,Q2] = svd(WI{2});
off_wip = toc; fprintf('#');

tic;
[P3,E3,Q3] = svd(WJ{2});
off_wjp = toc; fprintf('#');

tic;
[U7,D7,V7] = bal(WS{1},WO);
off_wss = toc; fprintf('#');

tic;
[U8,D8,V8] = bal(WC,WI{1});
off_wis = toc;

tic;
[U9,D9,V9] = svd(WJ{1});
off_wjs = toc; fprintf('#\n');


%%%%%%% Online %%%%%%%%

%POD
for I=1:N-1, y = assess(Fl,Gl,X,U,P,S,h,T,U1(:,1:I)',U1(:,1:I),speye(N),speye(N)); L2(1,I) = l2norm(Y-y); end; fprintf('#');

%WC
for I=1:N-1, y = assess(Fl,Gl,X,U,P,S,h,T,U2(:,1:I)',U2(:,1:I),speye(N),speye(N)); L2(2,I) = l2norm(Y-y); end; fprintf('#');

%WCWO
for I=1:N-1, y = assess(Fl,Gl,X,U,P,S,h,T,U3(1:I,:),V3(:,1:I),speye(N),speye(N)); L2(3,I) = l2norm(Y-y); end; fprintf('#');

%WCWQ
for I=1:N-1, y = assess(Fl,Gl,X,U,P,S,h,T,U4(1:I,:),V4(:,1:I),speye(N),speye(N)); L2(4,I) = l2norm(Y-y); end; fprintf('#');

%WX
for I=1:N-1, y = assess(Fl,Gl,X,U,P,S,h,T,U5(:,1:I)',U5(:,1:I),speye(N),speye(N)); L2(5,I) = l2norm(Y-y); end; fprintf('#');

%WY
for I=1:N-1, y = assess(Fl,Gl,X,U,P,S,h,T,U6(:,1:I)',U6(:,1:I),speye(N),speye(N)); L2(6,I) = l2norm(Y-y); end; fprintf('#');

%WS
for I=1:N-1, tic; y = assess(Fl,Gl,X,U,P,S,h,T,speye(N),speye(N),P1(:,1:I)',P1(:,1:I)); L2(7,I) = l2norm(Y-y); end; fprintf('#');

%WI
for I=1:N-1, tic; y = assess(Fl,Gl,X,U,P,S,h,T,speye(N),speye(N),P2(:,1:I)',P2(:,1:I)); L2(8,I) = l2norm(Y-y); end; fprintf('#');

%WII
for I=1:N-1, tic; y = assess(Fl,Gl,X,U,P,S,h,T,speye(N),speye(N),P3(:,1:I)',P3(:,1:I)); L2(9,I) = l2norm(Y-y); end; fprintf('#');

%WSWO
for I=1:n*n-1, for K=1:n*n-1, tic; y = assess(Fl,Gl,X,U,P,S,h,T,U7(1:I*n,:),V7(:,1:I*n),P1(:,1:K*n)',P1(:,1:K*n)); l2(I,K,1) = l2norm(Y-y); end; end; fprintf('#');

%WCWI
for I=1:n*n-1, for K=1:n*n-1, tic; y = assess(Fl,Gl,X,U,P,S,h,T,U8(1:I*n,:),V8(:,1:I*n),P2(:,1:K*n)',P2(:,1:K*n)); l2(I,K,2) = l2norm(Y-y); end; end; fprintf('#');

%WJ
for I=1:n*n-1, for K=1:n*n-1, tic; y = assess(Fl,Gl,X,U,P,S,h,T,U9(:,1:I*n)',U9(:,1:I*n),P3(:,1:K*n)',P3(:,1:K*n)); l2(I,K,3) = l2norm(Y-y); end; end; fprintf('#\n');


%%%%%%%% Plot %%%%%%%%

figure;
subplot(5,1,[1,4]);
semilogy(1:N,L2(1,:)./l2norm(Y),'r','LineWidth',2);
hold on;
semilogy(1:N,L2(2,:)./l2norm(Y),'b','LineWidth',2);
hold off;
xlim([1,N-1]); ylim([1e-16,1]);
legend('POD','W_C');
xlabel('State Dimension'); ylabel('Relative L^2 Output Error');
subplot(5,1,5);
barh([off_wc+off_wcc;0],'b'); hold on;
barh([0;off_wu+off_pod],'r'); hold off;
xlabel('Offline Time [s]'); set(gca,'YTickLabel',{'Wc','POD'});


figure;
subplot(5,1,[1,4]);
semilogy(1:N,L2(3,:)./l2norm(Y),'r','LineWidth',2);
hold on;
semilogy(1:N,L2(4,:)./l2norm(Y),'b','LineWidth',2);
hold off;
xlim([1,N-1]); ylim([1e-16,1]);
legend('W_C + W_O','W_C + W_C^*');
xlabel('State Dimension'); ylabel('Relative L^2 Output Error');
subplot(5,1,5);
barh([off_wc+off_wq+off_bt2;0],'b'); hold on;
barh([0;off_wc+off_wo+off_bt1],'r'); hold off;
xlabel('Offline Time [s]'); set(gca,'YTickLabel',{'Wc + Wc*','Wc + Wo'});

figure;
subplot(5,1,[1,4]);
semilogy(1:N,L2(5,:)./l2norm(Y),'r','LineWidth',2);
hold on;
semilogy(1:N,L2(6,:)./l2norm(Y),'b','LineWidth',2);
hold off;
xlim([1,N-1]); ylim([1e-16,1]);
legend('W_X','W_Y');
xlabel('State Dimension'); ylabel('Relative L^2 Output Error');
subplot(5,1,5);
barh([off_wy+off_wx2;0],'b'); hold on;
barh([0;off_wx+off_wx1],'r'); hold off;
xlabel('Offline Time [s]'); set(gca,'YTickLabel',{'Wy','Wx'});


figure;
subplot(5,1,[1,4]);
semilogy(1:N,L2(7,:)./l2norm(Y),'r','LineWidth',2);
hold on;
semilogy(1:N,L2(8,:)./l2norm(Y),'b','LineWidth',2);
semilogy(1:N,L2(9,:)./l2norm(Y),'k','LineWidth',2);
hold off;
xlim([1,N]); ylim([1e-6,1]);
legend('W_S','W_I','W_{II}');
xlabel('Parameter Dimension'); ylabel('Relative L^2 Output Error');
subplot(5,1,5);
barh([off_wj+off_wjp;0;0],'k'); hold on;
barh([0;off_wi+off_wip;0],'b'); 
barh([0;0;off_ws+off_wsp],'r'); hold off;
xlabel('Offline Time [s]'); set(gca,'YTickLabel',{'Wii','Wi','Ws'});


figure;
subplot(5,3,[1,4,7,10]);
h = surf(l2(:,:,1)./l2norm(Y)); view(150,30); set(h,'EdgeColor',[0 0 0]);
cmap = autumn(100); cmap = flipud(cmap); tmp = linspace(0,1,100)';
colormap(cmap); set(gca,'zscale','log');
set(gca,'ZMinorGrid','off');
xlim([1,n*n]); xlabel('Parameter Dimension');
ylim([1,n*n]); ylabel('State Dimension');
zlim([1e-5,1]); zlabel('Relative L^2 Output Error');

subplot(5,3,[2,5,8,11]);
h = surf(l2(:,:,2)./l2norm(Y)); view(150,30); set(h,'EdgeColor',[0 0 0]);
cmap = autumn(100); cmap = flipud(cmap); tmp = linspace(0,1,100)';
colormap(cmap); set(gca,'zscale','log');
set(gca,'ZMinorGrid','off');
xlim([1,n*n]); xlabel('Parameter Dimension');
ylim([1,n*n]); ylabel('State Dimension');
zlim([1e-5,1]); zlabel('Relative L^2 Output Error');

subplot(5,3,[3,6,9,12]);
h = surf(l2(:,:,3)./l2norm(Y)); view(150,30); set(h,'EdgeColor',[0 0 0]);
cmap = autumn(100); cmap = flipud(cmap); tmp = linspace(0,1,100)';
colormap(cmap); set(gca,'zscale','log');
set(gca,'ZMinorGrid','off');
xlim([1,n*n]); xlabel('Parameter Dimension');
ylim([1,n*n]); ylabel('State Dimension');
zlim([1e-5,1]); zlabel('Relative L^2 Output Error');

subplot(5,3,[13,14,15]);
barh([off_wj+off_wjp+off_wjs;0;0],'k'); hold on;
barh([0;off_wi+off_wip+off_wc+off_wis;0],'b'); 
barh([0;0;off_ws+off_wsp+off_wo+off_wss],'r'); hold off;
xlabel('Offline Time [s]'); set(gca,'YTickLabel',{'Wj','Wc+Wi','Ws+Wo'});

end

%%%%

function [U Y V] = bal(WC,WO)

    [L D l] = svd(WC); LC = L*diag(sqrt(diag(D)));
    [L D l] = svd(WO); LO = L*diag(sqrt(diag(D)));
    [X Y Z] = svd(LO'*LC);
    U = ( LO*X*diag(1.0./sqrt(diag(Y))) )';
    V =   LC*Z*diag(1.0./sqrt(diag(Y)));

end

function y = assess(F,G,X,U,P,S,h,T,UN,VN,UP,VP)

    p = UP*P;
    x = UN*X;

    f = @(x,u,p) UN*F(VN*x,u,VP*p);
    g = @(x,u,p)    G(VN*x,u,VP*p);

    y = ode(f,g,x,U,p,S,h,T);

end

function y = ode(f,g,x,u,p,S,h,E)

 T = (E-S)/h;
 y = zeros(numel(g(x,u(:,1),p)),T);

 for t=1:T
  x = x + h*f(x + (0.5*h)*f(x,u(:,t),p),u(:,t),p); %Improved Eulers Method
  y(:,t) = g(x,u(:,t),p);
 end

end 
